<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Stack\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Factory\Stack\File\FileStackIndexFactory;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseBinaryFile;
use Org\Mxchange\CoreFramework\Filesystem\File\BinaryFile;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Stack\BaseStacker;
use Org\Mxchange\CoreFramework\Traits\Index\IndexableTrait;
use Org\Mxchange\CoreFramework\Traits\Iterator\IteratorTrait;
use Org\Mxchange\CoreFramework\Utils\Arrays\ArrayUtils;
use Org\Mxchange\CoreFramework\Utils\Strings\StringUtils;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;
use \UnexpectedValueException;

/**
 * A general file-based stack class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseFileStack extends BaseStacker {
	// Load traits
	use IndexableTrait;
	use IteratorTrait;

	// Exception codes
	const EXCEPTION_BAD_MAGIC = 0xe100;

	/**
	 * Minimum block length
	 */
	private static $minimumBlockLength = 0;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 * @todo	To hard assertions here, better rewrite them to exceptions
	 * @throws	UnexpectedValueException	If header is not proper length
	 * @throws	InvalidMagicException	If a bad magic was found
	 */
	public function readStackHeader () {
		// First rewind to beginning as the header sits at the beginning ...
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: CALLED!');
		$this->getIteratorInstance()->rewind();

		// Get header size
		$headerSize = $this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize();

		// Then read it (see constructor for calculation)
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: headerSize=%d', $headerSize));
		$data = $this->getIteratorInstance()->getBinaryFileInstance()->read($headerSize);

		// Have all requested bytes been read?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Read %d bytes (%d wanted).', strlen($data), $this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize()));
		if (strlen($data) != $this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize()) {
			// Bad data length
			throw new UnexpectedValueException(sprintf('data(%d)=%s does not match iteratorInstance->headerSize=%d',
				strlen($data),
				$data,
				$this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize()
			));
		} elseif (empty(trim($data, chr(0)))) {
			// Empty header, file is freshly pre-allocated
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Empty file header detected - EXIT!');
			return;
		}

		// Last character must be the separator
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: data(-1)=%s', dechex(ord(substr($data, -1, 1)))));
		if (substr($data, -1, 1) !== chr(BinaryFile::SEPARATOR_HEADER_ENTRIES)) {
			// Not valid separator
			throw new UnexpectedValueException(sprintf('data=%s does not have separator=%s at the end.',
				$data,
				BinaryFile::SEPARATOR_HEADER_ENTRIES
			));
		}

		// Okay, then remove it
		$data = substr($data, 0, -1);

		// And update seek position
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->iteratorInstance->updateSeekPosition() ...');
		$this->getIteratorInstance()->getBinaryFileInstance()->updateSeekPosition();

		/*
		 * Now split it:
		 *
		 * 0 => magic
		 * 1 => total entries
		 * 2 => current seek position
		 */
		$header = explode(chr(BinaryFile::SEPARATOR_HEADER_DATA), $data);

		// Map numeric indexes to associative indexes
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: header()=%d', count($header)));
		$header = ArrayUtils::mapNumericKeysToAssociative($header, [
			BinaryFile::HEADER_NAME_MAGIC,
			BinaryFile::HEADER_NAME_TOTAL_ENTRIES,
			BinaryFile::HEADER_NAME_SEEK_POSITION,
		]);

		// Check if the array has only 3 elements
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: HEADER_STACK_ELEMENT_COUNT=%d,header(%d)=%s', BinaryFile::HEADER_STACK_ELEMENT_COUNT, count($header), print_r($header, true)));
		if (count($header) != BinaryFile::HEADER_STACK_ELEMENT_COUNT) {
			// Header array count is not expected
			throw new UnexpectedValueException(sprintf('data=%s has %d elements, but expected is %d',
				$data,
				count($header),
				BinaryFile::HEADER_STACK_ELEMENT_COUNT
			));
		} elseif ($header[BinaryFile::HEADER_NAME_MAGIC] != StackableFile::STACK_MAGIC) {
			// Bad magic
			throw new InvalidMagicException($data, self::EXCEPTION_BAD_MAGIC);
		} elseif (strlen($header[BinaryFile::HEADER_NAME_TOTAL_ENTRIES]) != BinaryFile::LENGTH_COUNT) {
			// Count length not valid
			throw new UnexpectedValueException(sprintf('header[%s](%d)=%s is not expected %d length',
				BinaryFile::HEADER_NAME_TOTAL_ENTRIES,
				strlen($header[BinaryFile::HEADER_NAME_TOTAL_ENTRIES]),
				$header[BinaryFile::HEADER_NAME_TOTAL_ENTRIES],
				BinaryFile::LENGTH_COUNT
			));
		} elseif (strlen($header[BinaryFile::HEADER_NAME_SEEK_POSITION]) != BinaryFile::LENGTH_POSITION) {
			// Position length not valid
			throw new UnexpectedValueException(sprintf('header[%s](%d)=%s is not expected %d length',
				BinaryFile::HEADER_NAME_SEEK_POSITION,
				strlen($header[BinaryFile::HEADER_NAME_SEEK_POSITION]),
				$header[BinaryFile::HEADER_NAME_SEEK_POSITION],
				BinaryFile::LENGTH_POSITION
			));
		}

		// Decode count and seek position
		$header[BinaryFile::HEADER_NAME_TOTAL_ENTRIES] = hex2bin($header[BinaryFile::HEADER_NAME_TOTAL_ENTRIES]);
		$header[BinaryFile::HEADER_NAME_SEEK_POSITION] = hex2bin($header[BinaryFile::HEADER_NAME_SEEK_POSITION]);

		// Set header here
		$this->getIteratorInstance()->getBinaryFileInstance()->setHeader($header);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: EXIT!');
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	public function flushFileHeader () {
		// Put all informations together
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: CALLED!');
		$header = sprintf('%s%s%s%s%s%s',
			// Magic
			StackableFile::STACK_MAGIC,

			// Separator magic<->count
			chr(BinaryFile::SEPARATOR_HEADER_DATA),

			// Padded total entries
			str_pad(StringUtils::dec2hex($this->getIteratorInstance()->getBinaryFileInstance()->getCounter()), BinaryFile::LENGTH_COUNT, '0', STR_PAD_LEFT),

			// Separator count<->seek position
			chr(BinaryFile::SEPARATOR_HEADER_DATA),

			// Padded seek position
			str_pad(StringUtils::dec2hex($this->getIteratorInstance()->getBinaryFileInstance()->getSeekPosition(), 2), BinaryFile::LENGTH_POSITION, '0', STR_PAD_LEFT),

			// Separator position<->entries
			chr(BinaryFile::SEPARATOR_HEADER_ENTRIES)
		);

		// Write it to disk (header is always at seek position 0)
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Calling this->iteratorInstance->writeAtPosition(0, header=%s) ...', $header));
		$this->getIteratorInstance()->getBinaryFileInstance()->writeAtPosition(0, $header);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: EXIT!');
	}

	/**
	 * Initializes this file-based stack.
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @param	$type		Type of this stack (e.g. url_source for URL sources)
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 * @todo	Currently the stack file is not cached, please implement a memory-handling class and if enough RAM is found, cache the whole stack file.
	 */
	protected function initFileStack (SplFileInfo $fileInfoInstance, string $type) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: fileInfoInstance[%s]=%s,type=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance, $type));
		if (empty($type)) {
			// Invalid parameter
			throw new InvalidArgumentException('Parameter "type" is empty');
		}

		// Get a stack file instance
		$stackInstance = ObjectFactory::createObjectByConfiguredName('stack_file_class', array($fileInfoInstance, $this));

		// Get iterator instance
		$iteratorInstance = ObjectFactory::createObjectByConfiguredName('file_iterator_class', array($stackInstance));

		// Set iterator here
		$this->setIteratorInstance($iteratorInstance);

		// Calculate header size
		$headerSize = (
			strlen(StackableFile::STACK_MAGIC) +
			strlen(chr(BinaryFile::SEPARATOR_HEADER_DATA)) +
			BinaryFile::LENGTH_COUNT +
			strlen(chr(BinaryFile::SEPARATOR_HEADER_DATA)) +
			BinaryFile::LENGTH_POSITION +
			strlen(chr(BinaryFile::SEPARATOR_HEADER_ENTRIES))
		);

		// Setting it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Setting headerSize=%d ...', $headerSize));
		$this->getIteratorInstance()->getBinaryFileInstance()->setHeaderSize($headerSize);

		// Init counters and gaps array
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->iteratorInstance->initCountersGapsArray() ...');
		$this->getIteratorInstance()->getBinaryFileInstance()->initCountersGapsArray();

		/*
		 * Get stack index instance. This can be used for faster
		 * "defragmentation" and startup.
		 */
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Creating index instance for fileInfoInstance[%s]=%s,type=%s ...', get_class($fileInfoInstance), $fileInfoInstance, $type));
		$indexInstance = FileStackIndexFactory::createFileStackIndexInstance($fileInfoInstance, $type);

		// And set it here
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: indexInstance=%s', $indexInstance->__toString()));
		$this->setIndexInstance($indexInstance);

		// Is the file's header initialized?
		if (!$this->getIteratorInstance()->getBinaryFileInstance()->isFileHeaderInitialized()) {
			// First pre-allocate a bit
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->iteratorInstance->preAllocateFile(file_stack) ...');
			$this->getIteratorInstance()->getBinaryFileInstance()->preAllocateFile('file_stack');

			// Then create file header
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: this->iteratorInstance->createFileHeader() ...');
			$this->getIteratorInstance()->getBinaryFileInstance()->createFileHeader();
		}

		// Load the file header
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->readStackHeader() ...');
		$this->readStackHeader();

		// Is the index loaded correctly and the stack file is just created?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->indexInstance->isIndexLoaded() ...');
		if (!$this->getIndexInstance()->isIndexLoaded()) {
			/*
			 * Something horrible has happened to the index as it should be
			 * loaded at this point. The stack's file structure then needs to
			 * be analyzed and the index rebuild.
			 */
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calling this->iteratorInstance->analyzeFileStructure() ...');
			$this->getIteratorInstance()->getBinaryFileInstance()->analyzeFileStructure();

			// Rebuild index from file
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Calling this->iteratorInstance->rebuildIndexFromStack(%s) ...', $this->__toString()));
			$this->getIndexInstance()->rebuildIndexFromStack($this);
		}

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: EXIT!');
	}

	/**
	 * Adds a value to given stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$value			Value to add to this stacker
	 * @return	void
	 * @throws	FullStackerException	If the stack is full
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	InvalidArgumentException	Not all variable types are wanted here
	 */
	protected function addValueToStack (string $stackerName, $value) {
		// Validate parameter
		/* PRINTR-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s,value[]=%s - CALLED!', $stackerName, gettype($value)));
		//* PRINTR-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s,value[%s]=%s', $stackerName, gettype($value), print_r($value, true)));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		} elseif ($this->isStackFull($stackerName)) {
			// Stacker is full
			throw new FullStackerException([$this, $stackerName, $value], self::EXCEPTION_STACKER_IS_FULL);
		} elseif (is_resource($value) || is_object($value)) {
			// Not wanted type
			throw new InvalidArgumentException(sprintf('value[]=%s is not supported', gettype($value)));
		}

		/*
		 * Now add the value to the file stack which returns gap position, a
		 * hash and length of the raw data.
		 */
		$data = $this->getIteratorInstance()->getBinaryFileInstance()->writeValueToFile($stackerName, $value);

		// Add the hash and gap position to the index
		//* PRINTR-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: data=%s', print_r($data, true));
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: Calling this->indexInstance->addHashedDataToIndex(%s,data()=%d) ...', $stackerName, count($data)));
		$this->getIndexInstance()->addHashedDataToIndex($stackerName, $data);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: EXIT!');
	}

	/**
	 * Get last value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If the stack is empty
	 */
	protected function getLastValue (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new BadMethodCallException([$this, $stackerName], self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now get the last value
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: value[]=%s - EXIT!', gettype($value)));
		return $value;
	}

	/**
	 * Get first value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If the stack is empty
	 */
	protected function getFirstValue (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new BadMethodCallException([$this, $stackerName], self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now get the first value
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: value[]=%s - EXIT!', gettype($value)));
		return $value;
	}

	/**
	 * "Pops" last entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "poped" from array
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If the stack is empty
	 */
	protected function popLast (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new BadMethodCallException([$this, $stackerName], self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * "Pops" first entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "shifted" from array
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If the named stacker is empty
	 */
	protected function popFirst (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		} elseif ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new BadMethodCallException([$this, $stackerName], self::EXCEPTION_STACKER_IS_EMPTY);
		}

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * Checks whether the given stack is full
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isFull			Whether the stack is full
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	protected function isStackFull (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		}

		// @TODO Please implement this, returning false
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		$isFull = false;

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: isFull=%d - EXIT!', intval($isFull)));
		return $isFull;
	}

	/**
	 * Checks whether the given stack is empty
	 *
	 * @param	$stackerName		Name of the stack
	 * @return	$isEmpty			Whether the stack is empty
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If given stack is missing
	 */
	public function isStackEmpty (string $stackerName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		if (empty($stackerName)) {
			// No empty stack name
			throw new InvalidArgumentException('Parameter "stackerName" is empty');
		}

		// So, is the stack empty?
		$isEmpty = (($this->getStackCount($stackerName)) == 0);

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: isEmpty=%d - EXIT!', intval($isEmpty)));
		return $isEmpty;
	}

	/**
	 * Checks whether the given stack is initialized (set in array $stackers)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isInitialized	Whether the stack is initialized
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function isStackInitialized (string $stackerName) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Determines whether the EOF has been reached
	 *
	 * @return	$isEndOfFileReached		Whether the EOF has been reached
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function isEndOfFileReached () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for size of given stack (array count)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$count			Size of stack (array count)
	 */
	public function getStackCount (string $stackerName) {
		// Now, simply return the found count value, this must be up-to-date then!
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackerName=%s - CALLED!', $stackerName));
		$count = $this->getIteratorInstance()->getBinaryFileInstance()->getCounter();

		// Return count
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: count=%d - EXIT!', $count));
		return $count;
	}

	/**
	 * Calculates minimum length for one entry/block
	 *
	 * @return	$length		Minimum length for one entry/block
	 */
	public function calculateMinimumBlockLength () {
		// Is the value "cached"?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: CALLED!');
		if (self::$minimumBlockLength == 0) {
			// Calulcate it
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: Calculating ...');
			self::$minimumBlockLength =
				// Length of entry group
				BinaryFile::LENGTH_GROUP + strlen(chr(BinaryFile::SEPARATOR_GROUP_HASH)) +
				// Hash + value
				self::getHashLength() + strlen(chr(BinaryFile::SEPARATOR_HASH_VALUE)) + 1 +
				// Final separator
				strlen(chr(BinaryFile::SEPARATOR_ENTRIES));
		}

		// Return it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: self::minimumBlockLength=%d - EXIT!', self::$minimumBlockLength));
		return self::$minimumBlockLength;
	}

	/**
	 * Initializes counter for valid entries, arrays for damaged entries and
	 * an array for gap seek positions. If you call this method on your own,
	 * please re-analyze the file structure. So you are better to call
	 * analyzeFileStructure() instead of this method.
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initCountersGapsArray () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for header size
	 *
	 * @return	$totalEntries	Size of file header
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getHeaderSize () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Setter for header size
	 *
	 * @param	$headerSize		Size of file header
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function setHeaderSize (int $headerSize) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for header array
	 *
	 * @return	$totalEntries	Size of file header
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getHeader () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Setter for header
	 *
	 * @param	$header		Array for a file header
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function setHeader (array $header) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Updates seekPosition attribute from file to avoid to much access on file.
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function updateSeekPosition () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for total entries
	 *
	 * @return	$totalEntries	Total entries in this file
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getCounter () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes data at given position
	 *
	 * @param	$seekPosition	Seek position
	 * @param	$data			Data to be written
	 * @param	$flushHeader	Whether to flush the header (default: flush)
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function writeData (int $seekPosition, string $data, bool $flushHeader = true) {
		// Not supported
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes at given position by seeking to it.
	 *
	 * @param	$seekPosition	Seek position in file
	 * @param	$dataStream		Data to be written
	 * @return	mixed			Number of writes bytes or false on error
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function writeAtPosition (int $seekPosition, string $dataStream) {
		// Not supported
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes given value to the file and returns a hash and gap position for it
	 *
	 * @param	$stackName	Group identifier
	 * @param	$value		Value to be added to the stack
	 * @return	$data		Hash and gap position
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function writeValueToFile (string $stackName, $value) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function searchNextGap (int $length) {
		// Not supported here
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getBinaryFileInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * "Getter" for file size
	 *
	 * @return	$fileSize	Size of currently loaded file
	 */
	public function getFileSize () {
		// Call iterator's method
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-FILE-STACK: CALLED!');
		$size = $this->getIteratorInstance()->getBinaryFileInstance()->getFileSize();

		// Return size
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: size=%d - EXIT!', $size));
		return $size;
	}

	/**
	 * Writes given raw data to the file and returns a gap position and length
	 *
	 * @param	$stackName	Group identifier
	 * @param	$hash		Hash from encoded value
	 * @param	$encoded	Encoded value to be written to the file
	 * @return	$data		Gap position and length of the raw data
	 */
	public function writeDataToFreeGap (string $stackName, string $hash, string $encoded) {
		// Raw data been written to the file
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackName=%s,hash=%s,encoded()=%d - CALLED!', $stackName, $hash, strlen($encoded)));
		$rawData = sprintf('%s%s%s%s%s',
			$stackName,
			BinaryFile::SEPARATOR_GROUP_HASH,
			hex2bin($hash),
			BinaryFile::SEPARATOR_HASH_VALUE,
			$encoded
		);

		// Search for next free gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackName=%s,hash=%s,rawData()=%d', $stackName, $hash, strlen($rawData)));
		$gapPosition = $this->getIteratorInstance()->getBinaryFileInstance()->searchNextGap(strlen($rawData));

		// Gap position cannot be smaller than header length + 1
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: gapPosition=%d', $gapPosition));
		if ($gapPosition <= $this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize()) {
			// Improper gap position
			throw new UnexpectedValueException(sprintf('gapPosition[%s]=%d is not larger than headerSize=%d',
				gettype($gapPosition),
				$gapPosition,
				$this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize()
			));
		}

		// Then write the data at that gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackName=%s,hash=%s,gapPosition=%s', $stackName, $hash, $gapPosition));
		$this->getIteratorInstance()->getBinaryFileInstance()->writeData($gapPosition, $rawData);

		// Return gap position, hash and length of raw data
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-FILE-STACK: stackName=%s,hash=%s,rawData()=%d - EXIT!', $stackName, $hash, strlen($rawData)));
		return [
			StackableFile::ARRAY_NAME_GAP_POSITION => $gapPosition,
			StackableFile::ARRAY_NAME_HASH         => $hash,
			StackableFile::ARRAY_NAME_DATA_LENGTH  => strlen($rawData),
		];
	}

}
